<?php

namespace Drupal\signageos\Plugin\DigitalSignagePlatform;

use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\digital_signage_device\DeviceInterface;
use Drupal\digital_signage_device\Entity\Device;
use Drupal\digital_signage_platform\PlatformPluginBase;
use Drupal\digital_signage_schedule\ScheduleSpec;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\GuzzleException;

/**
 * Plugin implementation of the digital_signage_platform.
 *
 * @DigitalSignagePlatform(
 *   id = "signageos",
 *   label = @Translation("signageOS"),
 *   description = @Translation("Integrates into signageOS platform.")
 * )
 */
class SignageOs extends PlatformPluginBase {

  public const BASE_URL = 'https://api.signageos.io/v1/';

  public const APPLET_NAME = 'Drupal';

  public const APPLET_VERSION = '1.0.0';

  public const FRONT_APPLET_VERSION = '4.0.0';

  public const DEFAULT_ORIENTATION = 'LANDSCAPE';

  public const DEFAULT_RESOLUTION = 'FULL_HD';

  public const RESOLUTIONS = [
    self::DEFAULT_RESOLUTION => [
      'width' => 1920,
      'height' => 1080,
    ],
    'HD_READY' => [
      'width' => 1080,
      'height' => 720,
    ],
  ];

  public const PORTRAIT_QUERY = 'PORTRAIT';

  /**
   * @var string
   */
  private $clientId;

  /**
   * @var string
   */
  private $clientSecret;

  /**
   * @var string
   */
  private $welcomeMsg;

  /**
   * @var string
   */
  private $bgColor;

  /**
   * @var string
   */
  private $fgColor;

  /**
   * {@inheritdoc}
   */
  public function init(): void {
    $config = $this->configFactory->get('signageos.settings');
    $this->clientId = $config->get('client_id');
    $this->clientSecret = $config->get('client_secret');
    $this->welcomeMsg = $config->get('welcome_msg');
    $this->bgColor = $config->get('bg_color');
    $this->fgColor = $config->get('fg_color');
  }

  /**
   * {@inheritdoc}
   */
  public function scheduleBaseFields(array &$fields): void {
    $fields['signageos_extid'] = BaseFieldDefinition::create('string')
      ->setRevisionable(FALSE)
      ->setTranslatable(FALSE)
      ->setLabel(t('External ID'))
      ->setDescription(t('The external ID of the schedule entity.'))
      ->setRequired(TRUE)
      ->setSetting('max_length', 255)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'string',
        'weight' => -4,
      ])
      ->setDisplayConfigurable('view', TRUE);
  }

  /**
   * {@inheritdoc}
   */
  public function getPlatformDevices(): array {
    $this->messenger->addStatus('Receiving signageOS devices');

    $deviceEntities = [];

    foreach ($this->request('device') as $device) {
      $resolution = $this->getCurrentDeviceResolution($device);

      if ($this->isPortraitOrientation($resolution)) {
        $width = self::RESOLUTIONS[$resolution['resolution']]['height'];
        $height = self::RESOLUTIONS[$resolution['resolution']]['width'];
      }
      else {
        $width = self::RESOLUTIONS[$resolution['resolution']]['width'];
        $height = self::RESOLUTIONS[$resolution['resolution']]['height'];
      }

      $deviceEntity = Device::create([
        'bundle' => $this->getPluginId(),
        'extid' => $device['uid'],
        'title' => $device['name'],
        'status' => TRUE,
        'description' => $device['name'],
        'width' => $width,
        'height' => $height,
      ]);
      $deviceEntities[] = $deviceEntity;

    }

    return $deviceEntities;
  }

  /**
   * {@inheritdoc}
   */
  public function pushSchedule(DeviceInterface $device, $debug = FALSE, $reload_css = FALSE, $reload_assets = FALSE): void {
    $at = $this->dateFormatter->format($this->time->getRequestTime(), 'html_datetime');
    $spec = new ScheduleSpec($device);
    $configuration = $spec->getSpec($debug, $reload_css, $reload_assets);
    $content = [];
    foreach ($configuration['items'] as $key => $item) {
      $uid = $item['entity']['type'] . '_' . $item['entity']['id'];
      if ($item['type'] === 'html' && !isset($content[$uid])) {
        $elements = [
          '#theme' => 'digital_signage_framework',
          '#content' => $this->renderer->buildEntityView(
            $item['entity']['type'],
            $item['entity']['id'],
            $device
          ),
          '#full_html' => FALSE,
        ];
        $content[$uid] = $this->renderer->renderPlain($elements);
      }
    }
    $configuration['content'] = $content;
    $this->request('timing', 'post', [
      'deviceUid' => $device->extId(),
      'appletUid' => $this->getAppletUid(),
      'appletVersion' => self::APPLET_VERSION,
      'configuration' => $configuration,
      'startsAt' => $at,
      'endsAt' => $at,
      'position' => 1,
      'finishEventType' => 'IDLE_TIMEOUT',
      'finishEventData' => 1234567890,
    ]);
    $this->request('device/' . $device->extId() . '/power-action', 'post', [
      'devicePowerAction' => 'APPLET_RELOAD',
    ]);
  }

  /**
   * @param $path
   * @param string $method
   * @param array $body
   *
   * @return array|string
   */
  protected function request($path, $method = 'get', $body = []) {
    $headers = [
      'Cache-Control' => 'no-cache',
      'Content-Type' => 'application/json',
      'X-Auth' => implode(':', [
        $this->clientId,
        $this->clientSecret,
      ]),
    ];
    $options['headers'] = $headers;
    if (!empty($body)) {
      $options['body'] = json_encode($body);
    }
    try {
      $client = new Client(['base_uri' => self::BASE_URL]);
      $response = $client->request($method, $path, $options);
      $statusCode = $response->getStatusCode();
      if ($statusCode === 200) {
        $content = $response->getBody()->getContents();
        return json_decode($content, TRUE);
      }
    }
    catch (GuzzleException $ex) {}
    return [];
  }

  /**
   * Gets the resolution of the given device.
   *
   * @param $device array|mixed|string
   *
   * @return array
   */
  protected function getCurrentDeviceResolution($device): array {
    $resolutions = $this->request('device/' . $device['uid'] . '/resolution');
    return $this->getLatestSucceededResolution($resolutions);
  }

  /**
   * Gets the latest resolution setting that was successful.
   *
   * @param $resolutions array
   *
   * @return array
   */
  protected function getLatestSucceededResolution($resolutions): array {
    $currentResolution = [
      'resolution' => self::DEFAULT_RESOLUTION,
      'orientation' => self::DEFAULT_ORIENTATION,
    ];
    $latestTimestamp = -1;
    foreach ($resolutions as $resolution) {
      if ($resolution['succeededAt'] === NULL) {
        continue;
      }
      if (strtotime($resolution['createdAt']) > $latestTimestamp) {
        $latestTimestamp = strtotime($resolution['succeededAt']);
        $currentResolution['resolution'] = $resolution['resolution'];
        $currentResolution['orientation'] = $resolution['orientation'];
      }
    }

    return $currentResolution;
  }

  /**
   * Gets whether the orientation is portrait or not.
   *
   * @param array $resolution
   *
   * @return bool
   */
  protected function isPortraitOrientation(array $resolution): bool {
    return strpos($resolution['orientation'], self::PORTRAIT_QUERY) === 0;
  }

  /**
   * @return string
   */
  protected function getAppletUidId(): string {
    return 'applet_uid';
  }

  /**
   * @return string
   */
  protected function getAppletName(): string {
    return self::APPLET_NAME . ': ' . $this->configFactory->get('system.site')->get('name');
  }

  /**
   * @param bool $push
   *
   * @return string|null
   */
  protected function getAppletUid($push = TRUE): ?string {
    $appletUid = $this->configFactory->get('signageos.settings')->get($this->getAppletUidId());
    if (empty($appletUid) && $push) {
      $appletUid = $this->pushApplet();
    }
    return $appletUid;
  }

  /**
   * @return array
   */
  protected function findOrCreateApplet() {
    foreach ($this->request('applet') as $applet) {
      if ($applet['name'] === $this->getAppletName()) {
        return $applet;
      }
    }
    $this->request('applet', 'post', ['name' => $this->getAppletName()]);
    return $this->findOrCreateApplet();
  }

  /**
   * @return string
   */
  public function pushApplet(): string {
    $logo_file = theme_get_setting('logo.url');
    if (pathinfo($logo_file, PATHINFO_EXTENSION) === 'svg') {
      $logo_content = file_get_contents(DRUPAL_ROOT . $logo_file);
      if ($pos = strpos($logo_content, '<svg')) {
        $logo_content = substr($logo_content, $pos);
      }
    }
    else {
      $logo_content = '';
    }

    // Build applet.
    $build = [
      '#theme' => 'signageos_applet',
      '#sitename' => $this->configFactory->get('system.site')->get('name'),
      '#welcome' => $this->welcomeMsg,
      '#bgcolor' => $this->bgColor,
      '#fgcolor' => $this->fgColor,
      '#logo' => $logo_content,
    ];
    $binary = (string) $this->renderer->renderPlain($build);

    // Find or create applet and version
    if ($applet_uid = $this->getAppletUid(FALSE)) {
      $applet = $this->request('applet/' . $applet_uid);
    }
    if (empty($applet)) {
      $applet = $this->findOrCreateApplet();
    }
    $appletRelease = $this->request('applet/' . $applet['uid'] . '/version/' . self::APPLET_VERSION);
    if (empty($appletRelease)) {
      // Create the applet release.
      $this->request('applet/' . $applet['uid'] . '/version', 'post', [
        'binary' => $binary,
        'version' => self::APPLET_VERSION,
        'frontAppletVersion' => self::FRONT_APPLET_VERSION,
      ]);
      $appletRelease = $this->request('applet/' . $applet['uid'] . '/version/' . self::APPLET_VERSION);
      $this->messenger->addStatus('Created new applet ' . $applet['name']);
    }
    else {
      // Update an applet release.
      $this->request('applet/' . $applet['uid'] . '/version/' . self::APPLET_VERSION, 'put', [
        'binary' => $binary,
        'frontAppletVersion' => self::FRONT_APPLET_VERSION,
      ]);
      $this->messenger->addStatus('Updated applet ' . $applet['name']);
    }

    if (empty($applet_uid) || $applet_uid !== $appletRelease['appletUid']) {
      $this->configFactory->getEditable('signageos.settings')
        ->set($this->getAppletUidId(), $appletRelease['appletUid'])
        ->save();
    }

    return $appletRelease['appletUid'];
  }

  public function debugDevice(DeviceInterface $device): void {
    $this->request('device/' . $device->extId() . '/debug', 'put', [
      'appletEnabled' => TRUE,
      'nativeEnabled' => TRUE,
    ]);
  }

  public function resolutionDevice(DeviceInterface $device, $resolution, $orientation): void {
    $this->request('device/' . $device->extId() . '/resolution', 'put', [
      'resolution' => $resolution,
      'orientation' => $orientation,
    ]);
  }

}
